import { Mapper } from '@automapper/core';
import { InjectMapper } from '@automapper/nestjs';
import { Controller, UsePipes } from '@nestjs/common';
import { CommandBus, QueryBus } from '@nestjs/cqrs';
import { GrpcMethod, RpcException } from '@nestjs/microservices';
import { DatabaseException } from 'src/modules/database/src/exceptions/database.exception';
import { AddUsernameCommand } from '../../commands/add-username.command';
import { CreateAuthenticationCommand } from '../../commands/create-authentication.command';
import { DeleteAuthenticationCommand } from '../../commands/delete-authentication.command';
import { DeleteUsernameCommand } from '../../commands/delete-username.command';
import { UpdatePasswordCommand } from '../../commands/update-password.command';
import { UpdateUsernameCommand } from '../../commands/update-username.command';
import { AddUsernameRequest } from '../../domain/dtos/add-username.request';
import { CreateAuthenticationRequest } from '../../domain/dtos/create-authentication.request';
import { DeleteAuthenticationRequest } from '../../domain/dtos/delete-authentication.request';
import { DeleteUsernameRequest } from '../../domain/dtos/delete-username.request';
import { UpdatePasswordRequest } from '../../domain/dtos/update-password.request';
import { UpdateUsernameRequest } from '../../domain/dtos/update-username.request';
import { ValidateAuthenticationRequest } from '../../domain/dtos/validate-authentication.request';
import { Authentication } from '../../domain/entities/authentication';
import { Username } from '../../domain/entities/username';
import { ValidateAuthenticationQuery } from '../../queries/validate-authentication.query';
import { AuthenticationPresenter } from './authentication.presenter';
import { RpcValidationPipe } from '../../../../utils/pipes/rpc.validation-pipe';
import { UsernamePresenter } from './username.presenter';

@UsePipes(
  new RpcValidationPipe({
    whitelist: true,
    forbidUnknownValues: false,
  }),
)
@Controller()
export class AuthenticationController {
  constructor(
    private readonly _commandBus: CommandBus,
    private readonly _queryBus: QueryBus,
    @InjectMapper() private readonly _mapper: Mapper,
  ) {}

  @GrpcMethod('AuthenticationService', 'Validate')
  async validate(
    data: ValidateAuthenticationRequest,
  ): Promise<AuthenticationPresenter> {
    try {
      const authentication: Authentication = await this._queryBus.execute(
        new ValidateAuthenticationQuery(data.username, data.password),
      );
      return this._mapper.map(
        authentication,
        Authentication,
        AuthenticationPresenter,
      );
    } catch (e) {
      throw new RpcException({
        code: 7,
        message: 'Permission denied',
      });
    }
  }

  @GrpcMethod('AuthenticationService', 'Create')
  async createUser(
    data: CreateAuthenticationRequest,
  ): Promise<AuthenticationPresenter> {
    try {
      const authentication: Authentication = await this._commandBus.execute(
        new CreateAuthenticationCommand(data),
      );
      return this._mapper.map(
        authentication,
        Authentication,
        AuthenticationPresenter,
      );
    } catch (e) {
      if (e instanceof DatabaseException) {
        if (e.message.includes('Already exists')) {
          throw new RpcException({
            code: 6,
            message: 'Auth already exists',
          });
        }
      }
      throw new RpcException({
        code: 7,
        message: 'Permission denied',
      });
    }
  }

  @GrpcMethod('AuthenticationService', 'AddUsername')
  async addUsername(data: AddUsernameRequest): Promise<UsernamePresenter> {
    try {
      const username: Username = await this._commandBus.execute(
        new AddUsernameCommand(data),
      );

      return this._mapper.map(username, Username, UsernamePresenter);
    } catch (e) {
      if (e instanceof DatabaseException) {
        if (e.message.includes('Already exists')) {
          throw new RpcException({
            code: 6,
            message: 'Username already exists',
          });
        }
      }
      throw new RpcException({
        code: 7,
        message: 'Permission denied',
      });
    }
  }

  @GrpcMethod('AuthenticationService', 'UpdateUsername')
  async updateUsername(
    data: UpdateUsernameRequest,
  ): Promise<UsernamePresenter> {
    try {
      const username: Username = await this._commandBus.execute(
        new UpdateUsernameCommand(data),
      );

      return this._mapper.map(username, Username, UsernamePresenter);
    } catch (e) {
      if (e instanceof DatabaseException) {
        if (e.message.includes('Already exists')) {
          throw new RpcException({
            code: 6,
            message: 'Username already exists',
          });
        }
      }
      throw new RpcException({
        code: 7,
        message: 'Permission denied',
      });
    }
  }

  @GrpcMethod('AuthenticationService', 'UpdatePassword')
  async updatePassword(
    data: UpdatePasswordRequest,
  ): Promise<AuthenticationPresenter> {
    try {
      const authentication: Authentication = await this._commandBus.execute(
        new UpdatePasswordCommand(data),
      );

      return this._mapper.map(
        authentication,
        Authentication,
        AuthenticationPresenter,
      );
    } catch (e) {
      throw new RpcException({
        code: 7,
        message: 'Permission denied',
      });
    }
  }

  @GrpcMethod('AuthenticationService', 'DeleteUsername')
  async deleteUsername(data: DeleteUsernameRequest) {
    try {
      return await this._commandBus.execute(new DeleteUsernameCommand(data));
    } catch (e) {
      throw new RpcException({
        code: 7,
        message: 'Permission denied',
      });
    }
  }

  @GrpcMethod('AuthenticationService', 'Delete')
  async deleteAuthentication(data: DeleteAuthenticationRequest) {
    try {
      return await this._commandBus.execute(
        new DeleteAuthenticationCommand(data),
      );
    } catch (e) {
      throw new RpcException({
        code: 7,
        message: 'Permission denied',
      });
    }
  }
}
