import { TestingModule, Test } from '@nestjs/testing';
import { DatabaseModule } from '../../../database/database.module';
import { PrismaService } from '../../../database/src/adapters/secondaries/prisma-service';
import { DatabaseException } from '../../../database/src/exceptions/database.exception';
import { v4 } from 'uuid';
import { Type } from '../../domain/dtos/type.enum';
import { UsernameRepository } from '../../adapters/secondaries/username.repository';
import { Username } from '../../domain/entities/username';

describe('UsernameRepository', () => {
  let prismaService: PrismaService;
  let usernameRepository: UsernameRepository;

  const createUsernames = async (nbToCreate = 10) => {
    for (let i = 0; i < nbToCreate; i++) {
      await prismaService.username.create({
        data: {
          uuid: v4(),
          username: `john.doe.${i}@email.com`,
          type: Type.EMAIL,
        },
      });
    }
  };

  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      imports: [DatabaseModule],
      providers: [UsernameRepository, PrismaService],
    }).compile();

    prismaService = module.get<PrismaService>(PrismaService);
    usernameRepository = module.get<UsernameRepository>(UsernameRepository);
  });

  afterAll(async () => {
    await prismaService.$disconnect();
  });

  beforeEach(async () => {
    await prismaService.username.deleteMany();
  });

  describe('findAll', () => {
    it('should return an empty data array', async () => {
      const res = await usernameRepository.findAll();
      expect(res).toEqual({
        data: [],
        total: 0,
      });
    });

    it('should return a data array with 8 usernames', async () => {
      await createUsernames(8);
      const usernames = await usernameRepository.findAll();
      expect(usernames.data.length).toBe(8);
      expect(usernames.total).toBe(8);
    });

    it('should return a data array limited to 10 usernames', async () => {
      await createUsernames(20);
      const usernames = await usernameRepository.findAll();
      expect(usernames.data.length).toBe(10);
      expect(usernames.total).toBe(20);
    });
  });

  describe('findOne', () => {
    it('should return a username with uuid and email', async () => {
      const usernameToFind = await prismaService.username.create({
        data: {
          uuid: v4(),
          username: 'john.doe@email.com',
          type: Type.EMAIL,
        },
      });

      const username = await usernameRepository.findOne({
        username: 'john.doe@email.com',
        type: Type.EMAIL,
      });
      expect(username.uuid).toBe(usernameToFind.uuid);
    });

    it('should return null', async () => {
      const username = await usernameRepository.findOne({
        username: 'jane.doe@email.com',
        type: Type.EMAIL,
      });
      expect(username).toBeNull();
    });
  });

  describe('create', () => {
    it('should create a username with an email', async () => {
      const beforeCount = await prismaService.username.count();

      const usernameToCreate: Username = new Username();
      usernameToCreate.uuid = v4();
      usernameToCreate.username = 'john.doe@email.com';
      usernameToCreate.type = Type.EMAIL;
      const username = await usernameRepository.create(usernameToCreate);

      const afterCount = await prismaService.username.count();

      expect(afterCount - beforeCount).toBe(1);
      expect(username.uuid).toBeDefined();
    });
    it('should create a username with a phone number', async () => {
      const beforeCount = await prismaService.username.count();

      const usernameToCreate: Username = new Username();
      usernameToCreate.uuid = v4();
      usernameToCreate.username = '+33611223344';
      usernameToCreate.type = Type.PHONE;
      const username = await usernameRepository.create(usernameToCreate);

      const afterCount = await prismaService.username.count();

      expect(afterCount - beforeCount).toBe(1);
      expect(username.uuid).toBeDefined();
    });
    it('should create a username with an email for an existing uuid', async () => {
      const beforeCount = await prismaService.username.count();

      const uuid = v4();

      const firstUsernameToCreate: Username = new Username();
      firstUsernameToCreate.uuid = uuid;
      firstUsernameToCreate.username = '+33611223344';
      firstUsernameToCreate.type = Type.PHONE;
      const firstUsername = await usernameRepository.create(
        firstUsernameToCreate,
      );

      const secondUsernameToCreate: Username = new Username();
      secondUsernameToCreate.uuid = uuid;
      secondUsernameToCreate.username = 'john.doe@email.com';
      secondUsernameToCreate.type = Type.EMAIL;
      const secondUsername = await usernameRepository.create(
        secondUsernameToCreate,
      );

      const afterCount = await prismaService.username.count();

      expect(afterCount - beforeCount).toBe(2);
      expect(firstUsername.uuid).toEqual(secondUsername.uuid);
    });
    it('should throw DatabaseException if username already exists for a given type', async () => {
      const uuid = v4();

      const firstUsernameToCreate: Username = new Username();
      firstUsernameToCreate.uuid = uuid;
      firstUsernameToCreate.username = 'john.doe@email.com';
      firstUsernameToCreate.type = Type.EMAIL;
      await usernameRepository.create(firstUsernameToCreate);

      const secondUsernameToCreate: Username = new Username();
      secondUsernameToCreate.uuid = uuid;
      secondUsernameToCreate.username = 'jane.doe@email.com';
      secondUsernameToCreate.type = Type.EMAIL;

      await expect(
        usernameRepository.create(secondUsernameToCreate),
      ).rejects.toBeInstanceOf(DatabaseException);
    });
  });

  describe('update', () => {
    it('should update username email', async () => {
      const usernameToUpdate = await prismaService.username.create({
        data: {
          uuid: v4(),
          username: `john.doe@email.com`,
          type: Type.EMAIL,
        },
      });

      const toUpdate: Username = new Username();
      toUpdate.username = 'jane.doe@email.com';
      const updatedUsername = await usernameRepository.updateWhere(
        {
          uuid_type: {
            uuid: usernameToUpdate.uuid,
            type: usernameToUpdate.type,
          },
        },
        {
          username: toUpdate.username,
        },
      );

      expect(updatedUsername.uuid).toBe(usernameToUpdate.uuid);
      expect(updatedUsername.username).toBe('jane.doe@email.com');
    });

    it('should update username phone', async () => {
      const usernameToUpdate = await prismaService.username.create({
        data: {
          uuid: v4(),
          username: `+33611223344`,
          type: Type.PHONE,
        },
      });

      const toUpdate: Username = new Username();
      toUpdate.username = '+33622334455';
      const updatedUsername = await usernameRepository.updateWhere(
        {
          uuid_type: {
            uuid: usernameToUpdate.uuid,
            type: usernameToUpdate.type,
          },
        },
        {
          username: toUpdate.username,
        },
      );

      expect(updatedUsername.uuid).toBe(usernameToUpdate.uuid);
      expect(updatedUsername.username).toBe('+33622334455');
    });
    it('should throw DatabaseException if email not found', async () => {
      const toUpdate: Username = new Username();
      toUpdate.username = 'jane.doe@email.com';

      await expect(
        usernameRepository.updateWhere(
          {
            uuid_type: {
              uuid: '544572be-11fb-4244-8235-587221fc9104',
              type: Type.EMAIL,
            },
          },
          {
            username: toUpdate.username,
          },
        ),
      ).rejects.toBeInstanceOf(DatabaseException);
    });
    it('should throw DatabaseException if phone not found', async () => {
      const toUpdate: Username = new Username();
      toUpdate.username = '+33611223344';

      await expect(
        usernameRepository.updateWhere(
          {
            uuid_type: {
              uuid: '544572be-11fb-4244-8235-587221fc9104',
              type: Type.PHONE,
            },
          },
          {
            username: toUpdate.username,
          },
        ),
      ).rejects.toBeInstanceOf(DatabaseException);
    });
  });

  describe('delete', () => {
    it('should delete a username', async () => {
      const usernameToRemove = await prismaService.username.create({
        data: {
          uuid: v4(),
          username: `+33611223344`,
          type: Type.PHONE,
        },
      });
      await usernameRepository.deleteMany({ uuid: usernameToRemove.uuid });

      const count = await prismaService.username.count();
      expect(count).toBe(0);
    });

    it('should throw DatabaseException', async () => {
      await expect(
        usernameRepository.delete('544572be-11fb-4244-8235-587221fc9104'),
      ).rejects.toBeInstanceOf(DatabaseException);
    });
  });
});
