import { classes } from '@automapper/classes';
import { AutomapperModule } from '@automapper/nestjs';
import { Test, TestingModule } from '@nestjs/testing';
import { AuthenticationRepository } from '../../adapters/secondaries/authentication.repository';
import { Authentication } from '../../domain/entities/authentication';
import * as bcrypt from 'bcrypt';
import { UpdatePasswordRequest } from '../../domain/dtos/update-password.request';
import { UpdatePasswordCommand } from '../../commands/update-password.command';
import { UpdatePasswordUseCase } from '../../domain/usecases/update-password.usecase';
import { LoggingMessager } from '../../adapters/secondaries/logging.messager';

const updatePasswordRequest: UpdatePasswordRequest =
  new UpdatePasswordRequest();
updatePasswordRequest.uuid = 'bb281075-1b98-4456-89d6-c643d3044a91';
updatePasswordRequest.password = 'John123';

const updatePasswordCommand: UpdatePasswordCommand = new UpdatePasswordCommand(
  updatePasswordRequest,
);

const mockAuthenticationRepository = {
  update: jest
    .fn()
    .mockResolvedValueOnce({
      uuid: updatePasswordRequest.uuid,
      password: bcrypt.hashSync(updatePasswordRequest.password, 10),
    })
    .mockImplementation(() => {
      throw new Error('Error');
    }),
};

const mockMessager = {
  publish: jest.fn().mockImplementation(),
};

describe('UpdatePasswordUseCase', () => {
  let updatePasswordUseCase: UpdatePasswordUseCase;

  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      imports: [AutomapperModule.forRoot({ strategyInitializer: classes() })],
      providers: [
        {
          provide: AuthenticationRepository,
          useValue: mockAuthenticationRepository,
        },
        {
          provide: LoggingMessager,
          useValue: mockMessager,
        },
        UpdatePasswordUseCase,
      ],
    }).compile();

    updatePasswordUseCase = module.get<UpdatePasswordUseCase>(
      UpdatePasswordUseCase,
    );
  });

  it('should be defined', () => {
    expect(updatePasswordUseCase).toBeDefined();
  });

  describe('execute', () => {
    it('should update an auth with an new encrypted password', async () => {
      const newAuth: Authentication = await updatePasswordUseCase.execute(
        updatePasswordCommand,
      );

      expect(
        bcrypt.compareSync(updatePasswordRequest.password, newAuth.password),
      ).toBeTruthy();
    });
    it('should throw an error if auth does not exist', async () => {
      await expect(
        updatePasswordUseCase.execute(updatePasswordCommand),
      ).rejects.toBeInstanceOf(Error);
    });
  });
});
