import { Injectable } from '@nestjs/common';
import { PrismaClientKnownRequestError } from '@prisma/client/runtime';
import { DatabaseException } from '../../exceptions/database.exception';
import { ICollection } from '../../interfaces/collection.interface';
import { IRepository } from '../../interfaces/repository.interface';
import { PrismaService } from './prisma-service';

/**
 * Child classes MUST redefined _model property with appropriate model name
 */
@Injectable()
export abstract class PrismaRepository<T> implements IRepository<T> {
  protected _model: string;

  constructor(protected readonly _prisma: PrismaService) {}

  async findAll(
    page = 1,
    perPage = 10,
    where?: any,
    include?: any,
  ): Promise<ICollection<T>> {
    const [data, total] = await this._prisma.$transaction([
      this._prisma[this._model].findMany({
        where,
        include,
        skip: (page - 1) * perPage,
        take: perPage,
      }),
      this._prisma[this._model].count({
        where,
      }),
    ]);
    return Promise.resolve({
      data,
      total,
    });
  }

  async findOneByUuid(uuid: string): Promise<T> {
    try {
      const entity = await this._prisma[this._model].findUnique({
        where: { uuid },
      });

      return entity;
    } catch (e) {
      if (e instanceof PrismaClientKnownRequestError) {
        throw new DatabaseException(
          PrismaClientKnownRequestError.name,
          e.code,
          e.message,
        );
      } else {
        throw new DatabaseException();
      }
    }
  }

  async findOne(where: any, include?: any): Promise<T> {
    try {
      const entity = await this._prisma[this._model].findFirst({
        where: where,
        include: include,
      });

      return entity;
    } catch (e) {
      if (e instanceof PrismaClientKnownRequestError) {
        throw new DatabaseException(PrismaClientKnownRequestError.name, e.code);
      } else {
        throw new DatabaseException();
      }
    }
  }

  async create(entity: Partial<T> | any, include?: any): Promise<T> {
    try {
      const res = await this._prisma[this._model].create({
        data: entity,
        include: include,
      });

      return res;
    } catch (e) {
      if (e instanceof PrismaClientKnownRequestError) {
        throw new DatabaseException(
          PrismaClientKnownRequestError.name,
          e.code,
          e.message,
        );
      } else {
        throw new DatabaseException();
      }
    }
  }

  async update(uuid: string, entity: Partial<T>): Promise<T> {
    try {
      const updatedEntity = await this._prisma[this._model].update({
        where: { uuid },
        data: entity,
      });
      return updatedEntity;
    } catch (e) {
      if (e instanceof PrismaClientKnownRequestError) {
        throw new DatabaseException(
          PrismaClientKnownRequestError.name,
          e.code,
          e.message,
        );
      } else {
        throw new DatabaseException();
      }
    }
  }

  async updateWhere(
    where: any,
    entity: Partial<T> | any,
    include?: any,
  ): Promise<T> {
    try {
      const updatedEntity = await this._prisma[this._model].update({
        where: where,
        data: entity,
        include: include,
      });

      return updatedEntity;
    } catch (e) {
      if (e instanceof PrismaClientKnownRequestError) {
        throw new DatabaseException(
          PrismaClientKnownRequestError.name,
          e.code,
          e.message,
        );
      } else {
        throw new DatabaseException();
      }
    }
  }

  async delete(uuid: string): Promise<T> {
    try {
      const entity = await this._prisma[this._model].delete({
        where: { uuid },
      });

      return entity;
    } catch (e) {
      if (e instanceof PrismaClientKnownRequestError) {
        throw new DatabaseException(
          PrismaClientKnownRequestError.name,
          e.code,
          e.message,
        );
      } else {
        throw new DatabaseException();
      }
    }
  }

  async deleteMany(where: any): Promise<void> {
    try {
      const entity = await this._prisma[this._model].deleteMany({
        where: where,
      });

      return entity;
    } catch (e) {
      if (e instanceof PrismaClientKnownRequestError) {
        throw new DatabaseException(
          PrismaClientKnownRequestError.name,
          e.code,
          e.message,
        );
      } else {
        throw new DatabaseException();
      }
    }
  }
}
